# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""
These are utility functions for the unit and integration tests.
"""

import pickle
import os

# pylint: disable=useless-suppression
# pylint: disable=unused-variable
# pylint: disable=pointless-statement


def export_pickle(filepath, obj, makedirs=True):
    """
    Auxiliary function to export a pickle object.
    Parameters
    ----------
    filepath: str
      The path of the file to be exported,
      including any subdirectories.
    obj: object
      The object to be pickled
    makedirs: bool
      If True, then the directories preceding the filename
      will be created if they do not exist.
    """
    # extract the directory name
    dirname = os.path.dirname(filepath)
    # if making directories is requested,
    if makedirs:
        # and the path does not exist
        if not os.path.exists(dirname):
            # create the directory
            os.makedirs(dirname)
    # open the file with the given filepath
    with open(filepath, 'wb') as f:
        # and store the object in the file
        pickle.dump(obj, f)


def import_pickle(filepath):
    """
    Auxiliary function to import a pickle object.
    Parameters
    ----------
    filepath: str
      The path of the file to be imported.

    Returns
    -------
    The pickled object.

    """
    # open the file with the given filepath
    with open(filepath, 'rb') as f:
        # and retrieve the pickled object
        return pickle.load(f)
